//  Copyright (C) 2012  Stephen Compall
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

namespace Nocandy.LinqResources.FS.ReaderV

open System
open Nocandy.LinqResources.FS.Mystery

type ReaderV<'r, 'e, 'a> =
    | ReaderV of ('r -> Mystery<'e, 'a>)

    /// Execute the reader until becoming aware of failure.
    member v.Run r =
        match v with ReaderV f -> f r

    /// Product of arrows.
    member v.Product(ov: ReaderV<'s, 'e, 'b>)
             : ReaderV<'r * 's, 'e, 'a * 'b> =
        ReaderV (fun (r, s) -> mystery.s { let! a = v.Run r
                                           let! b = ov.Run s
                                           return (a, b) })

    /// Change the input value type.
    member v.Contramap f = ReaderV (f >> v.Run)

    /// Change the output value type; forbids failure.
    member v.Map f = ReaderV (fun r -> mystery.s { let! a = v.Run r
                                                   return f a })

    /// Categorical composition.
    member v.Compose(fst: ReaderV<'s, 'e, 'r>): ReaderV<'s, 'e, 'a> =
        ReaderV (fun s -> mystery.s { let! r = fst.Run s
                                      return! v.Run r})

    /// Run the given effect first.
    member v.ButFirst (ef: ReaderV<'r, 'e, 'u>) =
        ef.Product(v).Contramap(fun s -> (s, s)).Map snd

    /// Fail if the condition fails.
    member v.Filter p =
        ReaderV(fun a -> mystery.s { if p a then return a }).Compose(v)

type ReaderVBuilder() =
    /// Supports let!, do!.
    member z.Bind(m: ReaderV<'r, 'e, 'a>,
                  f: 'a -> ReaderV<'r, 'e, 'b>): ReaderV<'r, 'e, 'b> =
        ReaderV (fun r -> mystery.s { let! a = m.Run r
                                      return! (f a).Run r })

    /// Supports let, do.
    member z.Let(x, f) : ReaderV<'r, 'e, 'a> =
        z.Delay (fun () -> f x)

    /// Supports delayed computation.
    member z.Delay (ffa: unit -> ReaderV<'r, 'e, 'a>) : ReaderV<'r, 'e, 'a> =
        ReaderV (fun r -> (ffa ()).Run r)

    /// Supports partial if.
    member z.Zero () = ReaderV (fun _ -> mystery.s.Zero())

    /// Supports lazy fallback.
    member z.Combine (a: ReaderV<'r, 'e, 'a>, b: ReaderV<'r, 'e, 'a>) =
        ReaderV (fun r -> let ar = a.Run r
                          ar.Fold((fun _ -> b.Run r),
                                  (fun _ -> b.Run r),
                                  (fun _ -> ar)))

    /// Supports return.
    member z.Return x = ReaderV (fun _ -> Success x)

    /// Supports return!.
    member z.ReturnFrom (fa: ReaderV<'r, 'e, 'a>) = fa

    /// Supports use.
    member z.Using (a: 'a, fb: 'a -> ReaderV<'r, 'e, 'b>)
        : ReaderV<'r, 'e, 'b> when 'a :> IDisposable =
        ReaderV (fun r -> use aa = a
                          (fb aa).Run r)

module rv =
    let s = ReaderVBuilder ()

    /// Categorical id: read the value.
    let ask = ReaderV (fun a -> mystery.s {return a})

    /// Arrow lifting; denies possibility of failure.
    let arr f = ReaderV (fun a -> mystery.s {return f a})

    /// Succeed or fail.
    let lift mys = ReaderV (fun _ -> mys)

    /// Distributive, or cosequence.
    let cosequence (xs: ReaderV<'r, 'e, 'a> seq): ReaderV<'r, 'e, 'a seq> =
        ReaderV (fun r -> xs |> Seq.map (fun x -> x.Run r)
                             |> mystery.sequence)
