//  Copyright (C) 2012  Stephen Compall
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

namespace Nocandy.LinqResources.FS.ReaderVExample

open Nocandy.LinqResources.FS.Mystery
open Nocandy.LinqResources.FS.ReaderV

type UserInfo = {Name : string; Age : int}

type Program = {User : UserInfo; Settings : string -> string}

module Scaffold =
    /// Succeed iff the program can be run.
    let mustBeRunnable() =
        rv.s { let! prg = rv.ask
               if prg.Settings "runnable" = "yes"
               then return () }

    /// Succeed iff the program can be run.  Like MustBeRunnable, but
    /// with a better error.
    let mustBeRunnable2 () =
        rv.s { let! prg = rv.ask
               if prg.Settings "runnable" = "yes"
               then return ()
               else return! "you must set runnable=yes"
                              |> UnknownError |> rv.lift }

    /// Whether the program should be run verbosely.
    let verbose () =
        rv.s { let! prg = rv.ask
               return prg.Settings "verbose" = "yes" }

    /// Describe the user in a long way.
    let describeLong () =
        rv.s { let! ui = rv.ask
               return sprintf "The user %s is of age %d" ui.Name ui.Age }

    /// Describe the user in a short way.
    let describeShort () =
        rv.s { let! ui = rv.ask
               return sprintf "%s is %d" ui.Name ui.Age }

    /// Describe the user, however the configuration says to.
    let describe () =
        rv.s { let! vb = verbose ()
               return! rv.s.Combine(describeLong().Filter(fun _ -> vb),
                                    describeShort())
                           .Contramap(fun p -> p.User) }

    let mapGet (m: Map<'k, 'v>) k = m.Item(k)

    let run (name, age, settings) =
        describe().ButFirst(mustBeRunnable2())
          .Run({User = {Name = name; Age = age};
                Settings = mapGet settings})
          .Fold(sprintf "ERROR: %s",
                sprintf "Exception: %A",
                id)
